from aim import Run
from omegaconf import DictConfig

from .device import set_device


class Action:
    def __init__(self, cfg: DictConfig, aim_run: Run):
        self.cfg = cfg
        self.action_cfg = cfg.action_config
        self.task_cfg = self.action_cfg.get("task")
        self.aim_run = aim_run
        set_device(self.action_cfg.device)



import sys
from importlib import import_module
from typing import Dict

from aim import Run


class ResumableSlurmJob:
    def __init__(self, module: str, action_name: str, cfg: Dict, aim_run: Run):
        self.module = module
        self.action_name = action_name
        self.cfg = cfg

        self.aim_run = None
        self.aim_run_hash = aim_run.hash

    def get_aim_run(self):
        if self.aim_run is None:
            self.aim_run = Run(self.aim_run_hash, repo=self.cfg.aim.repo)
        return self.aim_run

    def __call__(self):
        import submitit

        environment = submitit.JobEnvironment()
        master_ip = environment.hostnames[0]
        master_port = self.cfg.slurm.port_id
        self.cfg.slurm.init_method = "tcp"
        self.cfg.slurm.run_id = f"{master_ip}:{master_port}"

        self.get_aim_run()
        self.aim_run.set(
            "job", {"job_id": int(environment.job_id), "hostname": environment.hostname}
        )

        sys.path.append(f"{self.module}/actions")
        action = import_module(self.action_name)
        action.main(cfg=self.cfg, aim_run=self.aim_run)

    def checkpoint(self):
        import submitit

        runner = ResumableSlurmJob(
            module=self.module,
            action_name=self.action_name,
            cfg=self.cfg,
            aim_run=self.aim_run,
        )
        return submitit.helpers.DelayedSubmission(runner)

    def on_job_fail(self):
        self.get_aim_run()
        self.aim_run.close()


class ResumableJob:
    def __init__(self, module: str, action_name: str, cfg: Dict, aim_run: Run):
        self.module = module
        self.action_name = action_name
        self.cfg = cfg
        self.aim_run = aim_run

    def __call__(self):
        sys.path.append(f"{self.module}/actions")
        action = import_module(self.action_name)
        action.main(cfg=self.cfg, aim_run=self.aim_run)



import argparse
import logging
import re

from aim import Run
from hydra import compose, initialize

from .command import Command

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)

from ..utils.launcher import launch, launch_on_slurm
from ..utils.slurm import is_submitit_available
from ..utils.registry import Registry


@Command.register("launch")
class Launch(Command):
    """
    Launches an action from a specific module
    """

    def add_subparser(
        self, parser: argparse._SubParsersAction
    ) -> argparse.ArgumentParser:
        description = """urartu: launcher"""
        subparser = parser.add_parser(
            self.name,
            description=description,
            help="Launch an action from a given module",
        )

        subparser.add_argument("--name", type=str, help="name of the project/module")
        subparser.add_argument(
            "module_args", nargs=argparse.REMAINDER, help="module arguments"
        )

        subparser.set_defaults(fire=self._launch)

        return subparser

    def _launch(self, args: argparse.Namespace):
        module_name = re.sub(r"[^A-Za-z0-9]+", "", args.name)
        module_path = Registry.get_module_path_by_name(module_name)
 
        with initialize(version_base=None, config_path="../config"):
            cfg = compose(config_name="main", overrides=args.module_args)

        aim_run = Run(
            repo=cfg.aim.repo,
            experiment=cfg.action_config.experiment_name,
        )
        aim_run.set("cfg", cfg, strict=False)

        if cfg.slurm.use_slurm:
            assert (
                is_submitit_available()
            ), "Please 'pip install submitit' to schedule jobs on SLURM"

            launch_on_slurm(
                module=module_path,
                action_name=cfg.action_name,
                cfg=cfg,
                aim_run=aim_run,
            )
        else:
            launch(
                module=module_path,
                action_name=cfg.action_name,
                cfg=cfg,
                aim_run=aim_run,
            )

        if aim_run.active:
            aim_run.close()



import yaml
import os
from pathlib import Path

registry_file_path = Path(os.path.dirname(os.path.abspath(__file__))).parent.joinpath("registry.yaml")


class Registry:
    REGISTRY_FILE_PATH = registry_file_path

    @staticmethod
    def add_entry(module_name, module_root_dir):
        file_content = Registry.load_file_content()

        if module_name in file_content:
            raise ValueError(f"Module with name '{module_name}' already exists in the registery, use a different name")
        if str(module_root_dir) in file_content.values():
            raise ValueError(
                f"Module with path '{str(module_root_dir)}' already exists in the registery, use a different path"
            )
        if not module_root_dir.exists() or module_root_dir.is_file():
            raise FileNotFoundError(f"Couldn't find directory: {module_root_dir}")
        else:
            file_content[module_name] = str(module_root_dir)

            with open(Registry.REGISTRY_FILE_PATH, "a") as file:
                yaml.dump(file_content, file)

            return True

    @staticmethod
    def load_file_content():
        if not Registry.REGISTRY_FILE_PATH.exists():
            Registry.REGISTRY_FILE_PATH.touch()
            return {}

        with open(Registry.REGISTRY_FILE_PATH, "r") as file:
            file_content = yaml.safe_load(file)

        if file_content is None:
            raise RuntimeError(
                f"Registery is empty, register a module using `urartu register --name='NAME' --path='PATH'` command."
            )

        with open(Registry.REGISTRY_FILE_PATH, "r") as file:
            file_content = yaml.safe_load(file)

        return file_content

    @staticmethod
    def get_module_paths():
        file_content = Registry.load_file_content()

        return list(file_content.values())

    @staticmethod
    def get_module_path_by_name(module_name):
        file_content = Registry.load_file_content()

        if module_name not in file_content:
            raise KeyError(
                f"Registery with name '{module_name}' is not found, start with registering a module using `urartu register --name='NAME' --path='PATH'` command."
            )
        return file_content[module_name]



import logging

import torch

DEVICE = None


def set_device(device_name):
    if device_name == "auto":
        DEVICE = "auto"
    elif device_name == "cuda":
        assert device_name == "cuda" and torch.cuda.is_available()
        DEVICE = torch.device("cuda")
    else:
        DEVICE = torch.device("cpu")

    logging.info(f"Using DEVICE: {DEVICE}")



import time
from pathlib import Path
from typing import Dict

from aim import Run
from iopath.common.file_io import g_pathmgr

from .io import makedir
from .job import ResumableJob, ResumableSlurmJob


def create_submitit_executor(cfg: Dict):
    import submitit

    log_folder = Path(cfg.slurm.log_folder).joinpath(str(time.time()))
    makedir(log_folder)
    assert g_pathmgr.exists(
        log_folder
    ), f"Specified cfg.slurm.log_folder={log_folder} doesn't exist"
    assert cfg.slurm.partition, "slurm.PARTITION must be set when using slurm"

    executor = submitit.AutoExecutor(folder=log_folder)
    timeout_min = cfg.slurm.time_hours * 60 + cfg.slurm.time_minutes
    executor.update_parameters(
        name=cfg.slurm.name,
        slurm_comment=cfg.slurm.comment,
        slurm_partition=cfg.slurm.partition,
        slurm_account=cfg.slurm.account,
        slurm_constraint=cfg.slurm.constraint,
        timeout_min=timeout_min,
        nodes=cfg.slurm.num_nodes,
        cpus_per_task=cfg.slurm.num_cpu_per_proc * cfg.slurm.num_proc_per_node,
        tasks_per_node=cfg.slurm.num_proc_per_node,
        gpus_per_node=cfg.slurm.num_gpu_per_node,
        slurm_mem=f"{cfg.slurm.mem_gb}G",
        mem_gb=cfg.slurm.mem_gb,
        slurm_additional_parameters=cfg.slurm.additional_parameters,
    )
    return executor


def launch_on_slurm(module: str, action_name: str, cfg: Dict, aim_run: Run):
    executor = create_submitit_executor(cfg)
    trainer = ResumableSlurmJob(
        module=module, action_name=action_name, cfg=cfg, aim_run=aim_run
    )

    job = executor.submit(trainer)
    print(f"SUBMITTED: {job.job_id}")

    return job


def launch(module: str, action_name: str, cfg: Dict, aim_run: Run):
    trainer = ResumableJob(
        module=module, action_name=action_name, cfg=cfg, aim_run=aim_run
    )
    trainer()



import logging

from iopath.common.file_io import g_pathmgr


def makedir(dir_path):
    is_success = False
    try:
        if not g_pathmgr.exists(dir_path):
            g_pathmgr.mkdirs(dir_path)
        is_success = True
    except BaseException:
        logging.info(f"Error creating directory: {dir_path}")
    return is_success



import importlib
import os
import pwd


def get_current_user() -> str:
    return pwd.getpwuid(os.getuid()).pw_name


def get_user_config_module(file_path):
    config_path = f"configs.configs_{get_current_user}"
    config_path_hydra = config_path.replace(".", "/")
    config_file = file_path.split("/")[-1].split(".")[0]

    config_module = importlib.import_module(
        f"{config_path}.{config_file}", package=__name__
    )

    return config_module, config_path_hydra



from typing import Tuple, Union

import tiktoken
from langchain.schema import HumanMessage
from langchain_openai import AzureChatOpenAI
from transformers import AutoModelForCausalLM, AutoTokenizer

from ..common.model import Model


class OpenAIModel(Model):
    def __init__(self, cfg) -> None:
        super().__init__(cfg)

    def _load_model(self) -> Tuple[AutoModelForCausalLM, AutoTokenizer]:
        self.model = AzureChatOpenAI(
            deployment_name=self.cfg.name,
            openai_api_type=self.cfg.openai_api_type,
            openai_api_version=self.cfg.openai_api_version,
            azure_endpoint=self.cfg.azure_openai_endpoint,
            openai_api_key=self.cfg.azure_openai_api_key,
        )

    def generate(self, prompt: Union[str, Tuple[str, str]], generate_cfg):
        output = self.model(HumanMessage(content=prompt))

        return output

    def _get_num_tokens(self, string: str, encoding_name: str = "gpt-3.5-turbo") -> int:
        encoding = tiktoken.encoding_for_model(encoding_name)
        num_tokens = len(encoding.encode(string))
        return num_tokens



from typing import Any, Dict, List

import hydra

from .device import DEVICE


class Model:
    def __init__(self, cfg: List[Dict[str, Any]]):
        self.cfg = cfg
        self.model = None
        self.tokenizer = None
        self._load_model()

    @staticmethod
    def get_model(cfg):
        return hydra.utils.instantiate(cfg.type, cfg)

    def _load_model(self):
        raise NotImplementedError("method '_load_model' is not implemented")

    # def __call__(self, args):
    #     return self.model(**args)

    def generate(self, prompt):
        raise NotImplementedError("method 'generate' is not implemented")

    @staticmethod
    def collate_tokenize(data, tokenizer, input_key):
        input_batch = []
        for element in data:
            if isinstance(element[input_key], list):
                input_text = " ".join(element[input_key])
            else:
                input_text = element[input_key]
            input_batch.append(input_text)
        tokenized = tokenizer(
            input_batch, padding="longest", truncation=True, return_tensors="pt"
        ).to(DEVICE)
        return tokenized



import argparse
from typing import Callable, Dict, Optional, Type, TypeVar

from overrides import overrides

from ..utils.registrable import Registrable

T = TypeVar("T", bound="Command")


class Command(Registrable):

    _reverse_registry: Dict[Type, str] = {}

    def add_subparser(
        self, parser: argparse._SubParsersAction
    ) -> argparse.ArgumentParser:
        raise NotImplementedError

    @classmethod
    @overrides
    def register(
        cls: Type[T],
        name: str,
        constructor: Optional[str] = None,
        exist_ok: bool = False,
    ) -> Callable[[Type[T]], Type[T]]:
        super_register_fn = super().register(
            name, constructor=constructor, exist_ok=exist_ok
        )

        def add_name_to_reverse_registry(subclass: Type[T]) -> Type[T]:
            subclass = super_register_fn(subclass)
            # Don't need to check `exist_ok`, as it's done by super.
            # Also, don't need to delete previous entries if overridden, they can just stay there.
            cls._reverse_registry[subclass] = name
            return subclass

        return add_name_to_reverse_registry

    @property
    def name(self) -> str:
        return self._reverse_registry[self.__class__]



import argparse
import logging
import re
from pathlib import Path

from .command import Command
from ..utils.registry import Registry


@Command.register("register")
class Register(Command):
    """
    Registers a module by given location.
    """

    def add_subparser(self, parser: argparse._SubParsersAction) -> argparse.ArgumentParser:
        description = """urartu: register"""
        subparser = parser.add_parser(
            self.name,
            description=description,
            help="Register a new project in the registry",
        )

        subparser.add_argument("--name", type=str, help="name of the project/module")
        subparser.add_argument("--path", type=str, help="path of the root dir of the project/module")

        subparser.set_defaults(fire=self._register)

        return subparser

    def _register(self, args: argparse.Namespace):
        module_root_dir = Path(args.path)
        module_name = re.sub(r"[^A-Za-z0-9]+", "", args.name)

        if Registry.add_entry(module_name, module_root_dir):
            logging.info(f"Module '{module_name}' is registered successfully under path:'{module_root_dir}'")



import importlib
from collections import defaultdict
from typing import Callable, Dict, List, Optional, Tuple, Type, TypeVar

from .from_params import FromParams

T = TypeVar("T", bound="Registrable")


class Registrable(FromParams):
    _registry: Dict[Type, Dict[str, Tuple[Type, Optional[str]]]] = defaultdict(dict)
    default_implementation: Optional[str] = None

    @classmethod
    def register(
        cls: Type[T], name: str, constructor: str = None, exist_ok: bool = False
    ):

        registry = Registrable._registry[cls]

        def add_subclass_to_registry(subclass: Type[T]):
            # Add to registry, raise an error if key has already been used.
            if name in registry:
                if exist_ok:
                    message = (
                        f"{name} has already been registered as {registry[name][0].__name__}, but "
                        f"exist_ok=True, so overwriting with {cls.__name__}"
                    )
                    # logger.info(message)
                else:
                    message = (
                        f"Cannot register {name} as {cls.__name__}; "
                        f"name already in use for {registry[name][0].__name__}"
                    )
                    raise Exception(message)
            registry[name] = (subclass, constructor)
            return subclass

        return add_subclass_to_registry

    @classmethod
    def by_name(cls: Type[T], name: str) -> Callable[..., T]:
        """
        Returns a callable function that constructs an argument of the registered class.  Because
        you can register particular functions as constructors for specific names, this isn't
        necessarily the `__init__` method of some class.
        """
        # logger.debug(f"instantiating registered subclass {name} of {cls}")
        subclass, constructor = cls.resolve_class_name(name)
        if not constructor:
            return subclass
        else:
            return getattr(subclass, constructor)

    @classmethod
    def resolve_class_name(cls: Type[T], name: str) -> Tuple[Type[T], Optional[str]]:
        """
        Returns the subclass that corresponds to the given `name`, along with the name of the
        method that was registered as a constructor for that `name`, if any.

        This method also allows `name` to be a fully-specified module name, instead of a name that
        was already added to the `Registry`.  In that case, you cannot use a separate function as
        a constructor (as you need to call `cls.register()` in order to tell us what separate
        function to use).
        """
        if name in Registrable._registry[cls]:
            subclass, constructor = Registrable._registry[cls][name]
            return subclass, constructor
        elif "." in name:
            # This might be a fully qualified class name, so we'll try importing its "module"
            # and finding it there.
            parts = name.split(".")
            submodule = ".".join(parts[:-1])
            class_name = parts[-1]

            try:
                module = importlib.import_module(submodule)
            except ModuleNotFoundError:
                raise Exception(
                    f"tried to interpret {name} as a path to a class "
                    f"but unable to import module {submodule}"
                )

            try:
                subclass = getattr(module, class_name)
                constructor = None
                return subclass, constructor
            except AttributeError:
                raise Exception(
                    f"tried to interpret {name} as a path to a class "
                    f"but unable to find class {class_name} in {submodule}"
                )

        else:
            # is not a qualified class name
            raise Exception(
                f"{name} is not a registered name for {cls.__name__}. "
                "You probably need to use the --include-package flag "
                "to load your custom code. Alternatively, you can specify your choices "
                """using fully-qualified paths, e.g. {"model": "my_module.models.MyModel"} """
                "in which case they will be automatically imported correctly."
            )

    @classmethod
    def list_available(cls) -> List[str]:
        """List default first if it exists"""
        keys = list(Registrable._registry[cls].keys())
        default = cls.default_implementation

        if default is None:
            return keys
        elif default not in keys:
            raise Exception(f"Default implementation {default} is not registered")
        else:
            return [default] + [k for k in keys if k != default]



import argparse
from typing import Tuple

from .command import Command

from .launch import Launch
from .register import Register


def parse_args() -> Tuple[argparse.ArgumentParser, argparse.Namespace]:
    """
    Creates the argument parser for the main program and uses it to parse the args.
    """
    parser = argparse.ArgumentParser(description="Run urartu")

    subparsers = parser.add_subparsers(title="Commands", metavar="")

    def add_subcommands():
        for subcommand_name in sorted(Command.list_available()):
            subcommand_class = Command.by_name(subcommand_name)
            subcommand = subcommand_class()
            subcommand.add_subparser(subparsers)

    add_subcommands()
    args = parser.parse_args()

    return parser, args


def main():
    parser, args = parse_args()
    if "fire" in dir(args):
        args.fire(args)
    else:
        parser.print_help()



from typing import Any, Dict, List

from datasets import load_dataset

from ..common.dataset import Dataset


class HFDatasets(Dataset):
    def __init__(self, cfg: List[Dict[str, Any]]) -> None:
        super().__init__(cfg)

    def _get_dataset(self):
        self.dataset = load_dataset(
            self.cfg.name,
            self.cfg.get("subset"),
            split=self.cfg.get("split"),
        )



from typing import Tuple

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, pipeline

from ..common.device import DEVICE
from ..common.model import Model


class PipelineModel(Model):
    def __init__(self, cfg) -> None:
        super().__init__(cfg)

    def _load_model(self) -> Tuple[AutoModelForCausalLM, AutoTokenizer]:
        model = AutoModelForCausalLM.from_pretrained(
            self.cfg.name,
            cache_dir=self.cfg.get("cache_dir"),
            device_map=DEVICE,
            torch_dtype=eval(self.cfg.get("dtype")),
            token=self.cfg.get("api_token"),
        )
        self.tokenizer = AutoTokenizer.from_pretrained(self.cfg.name)

        self.model = pipeline(
            "text-generation",
            model=model,
            tokenizer=self.tokenizer,
            torch_dtype=eval(self.cfg.get("dtype")),
            device_map=DEVICE,
            eos_token_id=self.tokenizer.eos_token_id,
        )

    def generate(self, prompt: str, generate_cfg):
        output = self.model(prompt, **generate_cfg)
        output_first_resp = output[0]["generated_text"]

        return output_first_resp



from typing import Any, Dict, List

import hydra
from torch.utils.data import DataLoader

from .model import Model


class Dataset:
    def __init__(self, cfg: List[Dict[str, Any]]) -> None:
        self.cfg = cfg
        self.dataset = None
        self._get_dataset()

    @staticmethod
    def get_dataset(cfg):
        return hydra.utils.instantiate(cfg.type, cfg)

    def _get_dataset(self):
        raise NotImplementedError("method '_get_dataset' is not implemented")

    def get_dataloader(
        self, dataloader_cfg: List[Dict[str, Any]], tokenizer, input_key
    ):
        return DataLoader(
            self.dataset,
            batch_size=dataloader_cfg.get("batch_size"),
            num_workers=dataloader_cfg.get("num_workers", 2),
            shuffle=dataloader_cfg.get("shuffle"),
            collate_fn=lambda data: Model.collate_tokenize(
                data=data, tokenizer=tokenizer, input_key=input_key
            ),
        )



import os

from setuptools import find_packages, setup
from pathlib import Path

here = os.path.abspath(os.path.dirname(__file__))

# Version
version_file = os.path.join(here, "urartu", "VERSION")
with open(version_file) as vf:
    __version__ = vf.read().strip()

# Requirements
with open(os.path.join(here, "requirements.txt"), "r") as f:
    requirements = f.read().splitlines()

# Package info
NAME = "urartu"
DESCRIPTION = "An open-source NLP framework that offers high-level wrappers designed for effortless launch, enhanced reproducibility, superior control, and unmatched flexibility for your experiments."
VERSION = __version__
REQUIRES_PYTHON = ">=3.10.0"
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires=REQUIRES_PYTHON,
    install_requires=requirements,
    packages=find_packages("."),
    package_dir={"": "."},
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "urartu=urartu.__init__:main",
        ],
        "hydra.plugins": [
            "urartu_plugin = hydra_plugins.urartu_plugin.urartu_plugin:UrartuPlugin",
        ],
    },
)



import inspect
from typing import Callable, Generic, Type, TypeVar, Union

from .params import Params

T = TypeVar("T")


class Lazy(Generic[T]):
    """
    This class is for use when constructing objects using `FromParams`, when an argument to a
    constructor has a _sequential dependency_ with another argument to the same constructor.

    For example, in a `Trainer` class you might want to take a `Model` and an `Optimizer` as arguments,
    but the `Optimizer` needs to be constructed using the parameters from the `Model`. You can give
    the type annotation `Lazy[Optimizer]` to the optimizer argument, then inside the constructor
    call `optimizer.construct(parameters=model.parameters)`.

    This is only recommended for use when you have registered a `@classmethod` as the constructor
    for your class, instead of using `__init__`.  Having a `Lazy[]` type annotation on an argument
    to an `__init__` method makes your class completely dependent on being constructed using the
    `FromParams` pipeline, which is not a good idea.

    The actual implementation here is incredibly simple; the logic that handles the lazy
    construction is actually found in `FromParams`, where we have a special case for a `Lazy` type
    annotation.

    ```python
    @classmethod
    def my_constructor(
        cls,
        some_object: Lazy[MyObject],
        optional_object: Lazy[MyObject] = None,
        required_object_with_default: Lazy[MyObject] = Lazy(MyObjectDefault),
    ) -> MyClass:
        obj1 = some_object.construct()
        obj2 = None if optional_object is None else optional_object.construct()
        obj3 = required_object_with_default.construct()
    ```

    """

    def __init__(self, constructor: Union[Type[T], Callable[..., T]]):
        constructor_to_use: Callable[..., T]

        if inspect.isclass(constructor):

            def constructor_to_use(**kwargs):
                # type: ignore[union-attr]
                return constructor.from_params(Params({}), **kwargs)

        else:
            constructor_to_use = constructor

        self._constructor = constructor_to_use

    def construct(self, **kwargs) -> T:
        return self._constructor(**kwargs)



from typing import Tuple

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer

from ..common.device import DEVICE
from ..common.model import Model


class CausalLMModel(Model):
    def __init__(self, cfg) -> None:
        super().__init__(cfg)

    def _load_model(self) -> AutoModelForCausalLM:
        self.model = AutoModelForCausalLM.from_pretrained(
            self.cfg.name,
            cache_dir=self.cfg.get("cache_dir"),
            device_map=DEVICE,
            torch_dtype=eval(self.cfg.get("dtype")),
            token=self.cfg.get("api_token"),
        )

        for param in self.model.parameters():
            param.requires_grad = False

        self.tokenizer = AutoTokenizer.from_pretrained(self.cfg.name)
        if self.tokenizer.pad_token is None:
            self.tokenizer.pad_token = self.tokenizer.eos_token

    def generate(self, prompt: str, generate_cfg=None):
        if not generate_cfg:
            generate_cfg = self.cfg.get("generate")
        self.model.eval()

        prompt_tokenized = self.tokenizer(
            prompt, return_tensors="pt", padding=True, truncation=True
        )
        prompt_tensor = prompt_tokenized["input_ids"].to(self.model.device)
        attention_mask = prompt_tokenized["attention_mask"].to(self.model.device)
        with torch.no_grad():
            output_tokenized = self.model.generate(
                input_ids=prompt_tensor, attention_mask=attention_mask, **generate_cfg
            )
        output = self.tokenizer.decode(output_tokenized[0], skip_special_tokens=True)

        return output
